/*  game_centipede.h
 *
 *  Copyright (C) 2020-2026 Toxic All Rights Reserved.
 *
 *  This file is part of Toxic. Toxic is free software licensed
 *  under the GNU General Public License 3.0.
 */

#ifndef GAME_CENTIPEDE
#define GAME_CENTIPEDE

#include "game_base.h"

int centipede_initialize(GameData *game);

#endif  // GAME_CENTIPEDE
